@@ -18,6 +18,9 @@
 """
 A module for performing huffman encoding and decoding.
 
+0.1.2 Changelog:
+    More speedups
+
 0.1.1 Changelog:
     Fixed a problem when passing empty input file.
 """
@@ -36,43 +39,36 @@ except ImportError:
     from warnings import warn
     warn("Psyco not found, your decompression time will suffer with this.")
 
-__version__ = "0.1.1"
+__version__ = "0.1.2"
 __author__ = "Guilherme Polo <ggpolo@gmail.com>"
 
-def char_freq(file_obj, frequency_table):
+def char_freq(file_obj):
     """Build a frequency table from chars read from input file."""
     content = [ ]
     freq = defaultdict(int)
+
     for line in file_obj:
         content.append(line)
         for ch in line:
             freq[ch] += 1
 
-    # Create a leaf node for each item in freq
-    for ch, weight in freq.items():
-        frequency_table[ch] = weight
-
-    return content
+    # Return content read and create a leaf node for each item in freq
+    return content, HuffStructure(((ch, w) for ch, w in freq.items()))
 
 
-def read_header(f_in, frequency_table):
+def read_header(f_in):#, frequency_table):
     """Reads a huffman header."""
     header = f_in.read(1024) # 256 symbols, 4 bytes each
-
-    # unpack the same way it was packed, and build a frequency table
     unpacked_header = struct.unpack('256I', header)
-    for ch_indx, weight in enumerate(unpacked_header):
-        # several symbols with 0 weight were possibly stored, but we do
-        # not need them
-        if weight:
-            frequency_table[chr(ch_indx)] = weight
+
+    return HuffStructure(((chr(ch_indx), weight) for ch_indx, \
+                           weight in enumerate(unpacked_header) if weight))
 
 
-def write_header(f_out_obj, frequency_table):
+def write_header(f_out_obj, freq_table):
     """Write to file output a huffman header."""
-    header = (struct.pack('I', frequency_table[chr(i)].weight) \
-                for i in xrange(256))
-    f_out_obj.write(''.join(header))
+    f_out_obj.write(''.join([struct.pack('I', freq_table[chr(i)].weight) \
+                             for i in xrange(256)]))
 
 
 def codify_to_huffman(f_in_content, f_out_obj, codes):
@@ -81,57 +77,60 @@ def codify_to_huffman(f_in_content, f_out_obj, codes):
     input data."""
     bstr = ''.join([codes[ch] for line in f_in_content for ch in line])
     bstr_len_by8, remaining = divmod(len(bstr), 8)
-
-    bstrs = [bstr[i << 3:(i + 1) << 3] for i in xrange(bstr_len_by8)]
-
-    obstrs = [int(obstr, 2) for obstr in bstrs]
+    bstrs = [int(bstr[i << 3:(i + 1) << 3], 2) for i in xrange(bstr_len_by8)]
 
     if remaining:
-        # we need to complete last byte, so make sure it is a good 8-bit 
-        # string
-        obstrs.append(int(bstr[-remaining:].ljust(8, '0'), 2))
+        bstrs.append(int(bstr[-remaining:].ljust(8, '0'), 2))
 
     # write the code that identifies end of huffman encoding
     # this is actually the last byte in the file.
-    obstrs.append((8 - remaining) % 8) # last byte code
-    f_out_obj.write(struct.pack('%dB' % len(obstrs), *obstrs))
+    bstrs.append((8 - remaining) % 8) # last byte code
+
+    f_out_obj.write(struct.pack('%dB' % len(bstrs), *bstrs))
 
 
 def dec2bin(dec_list):
-    return array('c', ['1' if ((1 << i) & dec) else '0' for dec in dec_list \
-                                for i in xrange(7, -1, -1)])
+    """
+    Converts a sequence of decimal numbers to binary representation
+    and return a big string of 0's and 1's.
+    """
+    # calculate and store binary codes for the 256 symbols
+    bincodes = dict((k, ''.join(['1' if ((1 << i) & k) else '0' \
+                           for i in xrange(7, -1, -1)])) for k in xrange(256))
 
+    return ''.join([bincodes[dec] for dec in dec_list])
 
 def read_encoded(file_obj):
     """Reads a file compressed by huffman algorithm."""
     content = [ch for line in file_obj for ch in line]
-
     bstrs = dec2bin(struct.unpack('%dB' % len(content), ''.join(content)))
+    bstrs = bstrs[:-8 -int(bstrs[-8:], 2)]
 
-    end_code = int(bstrs[-8:].tostring(), 2)
-    
-    del bstrs[- 8 - end_code:] # remove end code and extra bytes from string
-    return bstrs.tostring()
+    return bstrs
 
 
 def decode_huffman(f_out_obj, root, binstr):
     """Expects a file object and a binstr that is returned from 
     read_encoded."""
+    w = f_out_obj.write
     start = root
+   
     for b in binstr:
         root = root.left if b == '0' else root.right
 
         if root.left is None and root.right is None:
             # found a leaf node, you found a symbol then
-            f_out_obj.write(root.data)
+            w(root.data)
             root = start
 
+
 class NoSymbols(Exception):
     """Raised when trying to build a tree for an empty input file."""
 
 
 class HuffNode(object):
     """Representation of a node in a 'Huffman tree'."""
+    #__slots__ = ('weight', 'data', 'left', 'right')
 
     def __init__(self, weight, data, left=None, right=None):
         self.weight = weight
@@ -146,7 +145,9 @@ class HuffNode(object):
 
 class HuffStructure(dict):
     """Data structure used in this Huffman implementation."""
-    def __init__(self):
+    def __init__(self, *args, **kwargs):
+        self.update(*args, **kwargs)
+
         self.codes = { } # symbols codes
         self.tree_built = False
 
@@ -182,6 +183,16 @@ class HuffStructure(dict):
         dict.__delitem__(self, key)
         self.cached = False
 
+    def update(self, *args, **kwargs):
+        """
+        Custom dict update, excepts args[0] to be an iterable with
+        values (x, y).
+        """
+        for k, v in args[0]:
+            self[k] = v
+        
+        self.cached = False
+
     def ordered_items(self):
         """Return items ordered by weight, so it is easier and faster
         to build the tree later."""
@@ -258,7 +269,6 @@ class HuffBase(object):
         self.file_out = open(file_out, out_mode)
 
         self.root = None
-        self.freq_table = HuffStructure()
 
     def gen_codes(self):
         """Build tree and generate codes for current nodes on it."""
@@ -274,7 +284,7 @@ class Huff(HuffBase):
     def __init__(self, file_in, file_out):
         HuffBase.__init__(self, file_in, file_out)
 
-        self.content = char_freq(self.file_in, self.freq_table)
+        self.content, self.freq_table = char_freq(self.file_in)
         write_header(self.file_out, self.freq_table)
         self.gen_codes()
 
@@ -290,7 +300,7 @@ class Unhuff(HuffBase):
     def __init__(self, file_in, file_out):
         HuffBase.__init__(self, file_in, file_out, False)
         
-        read_header(self.file_in, self.freq_table)
+        self.freq_table = read_header(self.file_in)
         self.gen_codes()
         self.decode_input()
         self.cleanup()
